@@ -54,9 +54,18 @@ instance Ord Number where
     (<) = binop (<) (<)
     {-# INLINE (<) #-}
 
+    (<=) = binop (<=) (<=)
+    {-# INLINE (<=) #-}
+
     (>) = binop (>) (>)
     {-# INLINE (>) #-}
 
+    (>=) = binop (>=) (>=)
+    {-# INLINE (>=) #-}
+
+    compare = binop compare compare
+    {-# INLINE compare #-}
+
 instance Num Number where
     (+) = binop (((I$!).) . (+)) (((D$!).) . (+))
     {-# INLINE (+) #-}
